unit WhatIsThis;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, MouseHook,
  ExtCtrls, CallWndProcHook, KeyboardHook;

type
  THelpShowMode = (hsmContext, hsmContextPopup);
  TWhatIsThisEvent = procedure (Sender : TObject; Control : TControl;
                                HelpContext : THelpContext; var ShowHelp : Boolean;
                                var ShowMode : THelpShowMode) of object;
  TNoHelpContextEvent = procedure (Sender : TObject; Control : TControl) of object;
  TWhatIsThis = class(TComponent)
  private
    FMouseHook : TMouseHook;
    FActive : Boolean;
    FOnWhatIsThis : TWhatIsThisEvent;
    FOnNoHelpContext : TNoHelpContextEvent;
    FOnActiveChanged : TNotifyEvent;
    FMonitorContextHelpButton : Boolean;
    WaitingButtonUp : Boolean;
    timer : TTimer;
    FCallWndProcHook : TCallWndProcHook;
    FKeyBoardHook : TKeyboardHook;
    ContextClicked : Boolean;
    FHelpShowMode : THelpShowMode;
    FMonitorF1 : Boolean;
    OldOnHelp : THelpEvent;
    ACallHelp : Boolean;
    CallCount : Integer;
    AppHelpFile : string;
    function MouseHookMouseProc(Sender: TObject; nCode: Integer;
                                wParam: Word; lParam: Integer): Integer;
    function CallWndProcHookProc (Sender: TObject; nCode: Integer;
                                  wParam: Word; lParam: Integer): Integer;
    function KeyboardHookProc (Sender: TObject; nCode: Integer;
                               wParam: Word; lParam: Integer): Integer;
    procedure SetActive (Value : Boolean);
    procedure OnTimer (Sender : TObject);
    procedure SetMonitorContextHelpButton (Value : Boolean);
    procedure SetMonitorF1 (Value : Boolean);
    procedure DoShowHelp (Control : TControl; hc : THelpContext);
    function OwnerOnHelp (Command: Word; Data: Longint; var CallHelp: Boolean): Boolean;
    function GetHelpContext (Control : TControl) : THelpContext;
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    property Active : Boolean read FActive Write SetActive;
  published
    property HelpShowMode : THelpShowMode read FHelpShowMode Write FHelpShowMode default hsmContext;
    property MonitorContextHelpButton : Boolean read FMonitorContextHelpButton Write SetMonitorContextHelpButton default false;
    property MonitorF1 : Boolean read FMonitorF1 Write SetMonitorF1 default false;
    property OnWhatIsThis : TWhatIsThisEvent read FOnWhatIsThis Write FOnWhatIsthis;
    property OnNoHelpContext : TNoHelpContextEvent read FOnNoHelpContext Write FOnNoHelpContext;
    property OnActiveChanged : TNotifyEvent read FOnActiveChanged Write FOnActiveChanged;
  end;

procedure Register;

implementation

uses
  typInfo;

constructor TWhatIsThis.Create;
begin
  inherited;
  FHelpShowMode := hsmContext;
  FCallWndProcHook := TCallWndProcHook.Create (Self);
  FCallWndProcHook.OnHookProc := CallWndProcHookProc;
  FMouseHook := TMouseHook.Create (Self);
  FMouseHook.OnHookProc := MouseHookMouseProc;
  FKeyBoardHook := TKeyboardHook.Create (Self);
  FKeyBoardHook.OnHookProc := KeyboardHookProc;
  timer := TTimer.Create (Self);
  with timer do
    begin
      Interval := 100;
      Enabled := False;
      OnTimer := Self.OnTimer;
    end;
  if not (csDesigning in ComponentState)
    then
    begin
      FMouseHook.Hooked := True;
      AppHelpFile := Application.HelpFile;
    end;  
end;

destructor TWhatIsThis.Destroy;
begin
  FMouseHook.Free;
  FCallWndProcHook.Free;
  FKeyBoardHook.Free;
  Timer.Free;
  inherited;
end;

function TWhatIsThis.MouseHookMouseProc;
var
  Control : TControl;
  function GetSubControl (AControl : TWinControl; p : TPoint) : TControl;
  var
    SubControl : TControl;
    p2 : TPoint;
  begin
    p2 := AControl.ScreenToClient (p);
    SubControl := AControl.ControlAtPos (p2, True, True);
    if SubControl <> nil
      then if SubControl is TWinControl
        then Result := GetSubControl (SubControl as TWinControl, p)
        else Result := SubControl
      else Result := AControl;
  end;
begin
  if FActive
    then
    begin
      Result := 1;
      if not WaitingButtonUp
        then if wParam = WM_LBUTTONDOWN
          then
          begin
            Control := FindControl (FMouseHook.MouseHookStruct.hwnd);
            if Control <> nil
              then
              begin
                with FMouseHook.MouseHookStruct.pt do
                  Control := GetSubControl (Control as TWinControl, Point (x, y));
                DoShowHelp (Control, 0);
              end;
            WaitingButtonUp := True;
            Active := False;
            timer.Enabled := True;
          end;
    end
    else if WaitingButtonUp
      then if wParam = WM_LBUTTONUP
        then
        begin
          WaitingButtonUp := False;
          Result := 1;
          FMouseHook.Hooked := False;
        end
        else Result := 1
      else Result := 0;
end;

procedure TWhatIsThis.SetActive;
begin
  if Value <> FActive
    then
    begin
      FActive := Value;
      if not FMonitorF1
        then FKeyBoardHook.Hooked := Value;
      if FActive
        then
        begin
          Screen.Cursor := crHelp;
          if not FMouseHook.Hooked
            then FMouseHook.Hooked := true
        end
        else
        begin
          Screen.Cursor := crDefault;
          ContextClicked := False;
        end;
      if Assigned (FOnActiveChanged)
        then FOnActiveChanged (Self);
    end;
end;

procedure TWhatIsThis.OnTimer;
begin
  timer.Enabled := False;
  if WaitingButtonUp
    then
    begin
      WaitingButtonUp := False;
      FMouseHook.Hooked := False;
    end;
end;

procedure TWhatIsThis.SetMonitorContextHelpButton;
begin
  if Value <> FMonitorContextHelpButton
    then
    begin
      FMonitorContextHelpButton := Value;
      if not (csDesigning in ComponentState)
        then FCallWndProcHook.Hooked := FMonitorContextHelpButton;
    end;
end;

function TWhatIsThis.CallWndProcHookProc;
begin
  Result := 0;
  with FCallWndProcHook.CWPStruct do
    if (Message = WM_SYSCOMMAND) and (wParam = SC_CONTEXTHELP) and
       ((Owner is TCustomForm) and ((Owner as TcustomForm).Handle = hwnd))
      then
      begin
        Active := True;
        Result := 1;
        ContextClicked := True;
      end
end;

function TWhatIsThis.GetHelpContext (Control : TControl) : THelpContext;
var
  PropInfo : PPropInfo;
begin
  PropInfo := GetPropInfo (Control, 'HelpContext');
  if PropInfo <> nil
    then
    begin
      Result := GetOrdProp (Control, PropInfo);
      if (Result = 0) and (Control.Parent <> nil)
        then Result := GetHelpContext (Control.Parent);
    end
    else if Control.Parent <> nil
      then Result := GetHelpContext (Control.Parent)
      else Result := 0;
end;

function TWhatIsThis.KeyboardHookProc;
var
  hc : THelpContext;
begin
  Result := 0;
  case wParam of
    VK_ESCAPE : if Active
      then
      begin
        if ContextClicked
          then Result := 0
          else Result := 1;
        Active := False;
      end;
    VK_F1 : if FMonitorF1 and (Owner is TForm) and (Screen.ActiveForm = Owner)
      then
      begin
        Result := 1;
        if Owner is TCustomForm
          then with (Owner as TCustomForm) do
            if ActiveControl <> nil
              then
              begin
                hc := GetHelpContext (ActiveControl);
                if hc <> 0
                  then
                  begin
                    CallCount := 2;
                    DoShowHelp (ActiveControl, hc);
                  end;
              end;
      end
  end;
end;

procedure TWhatIsThis.DoShowHelp;
var
  ShowHelp : Boolean;
  HelpContext : THelpContext;
  AHelpShowMode : THelpShowMode;
begin
  if hc = 0
    then HelpContext := GetHelpContext (Control)
    else HelpContext := hc;
  if HelpContext <> 0
    then
    begin
      ShowHelp := True;
      AHelpShowMode := FHelpShowMode;
      if Assigned (FOnWhatIsThis)
        then FOnWhatIsThis (Self, Control, HelpContext, ShowHelp, AHelpShowMode);
      if ShowHelp
        then
        begin
          ACallHelp := True;
          try
            case AHelpShowMode of
              hsmContext : Application.HelpContext (HelpContext);
              hsmContextPopup : Application.HelpCommand (HELP_CONTEXTPOPUP, HelpContext);
            end;
          finally
            ACallHelp := False;
          end;  
        end;
    end
    else if Assigned (FOnNoHelpContext)
      then FOnNoHelpContext (Self, Control);
end;

procedure TWhatIsThis.SetMonitorF1;
begin
  if Value <> FMonitorF1
    then
    begin
      FMonitorF1 := Value;
      if not (csDesigning in ComponentState)
        then
        begin
          FKeyBoardHook.Hooked := FMonitorF1;
          if Owner is TForm
            then with Owner as TForm do
              if FMonitorF1
                then
                begin
                  OldOnHelp := OnHelp;
                  OnHelp := OwnerOnHelp;
                end
                else OnHelp := OldOnHelp;
        end;
    end;
end;

function TWhatIsThis.OwnerOnHelp;
begin
  if Command in [HELP_CONTEXT, HELP_CONTEXTPOPUP]
    then
    begin
      with Owner as TForm do
        if HelpFile <> ''
          then Application.HelpFile := HelpFile
          else Application.HelpFile := AppHelpFile;
      CallHelp := (CallCount in [0, 2]) or ACallHelp;
      if CallCount > 0
        then Dec (CallCount);
      Result := true;
      if CallHelp
        then if Assigned (Application.OnHelp)
          then Result := Application.OnHelp (Command, Data, CallHelp);
    end
    else
    begin
      CallHelp := True;
      Result := True;
    end;
end;

{ Register }

procedure Register;
begin
  RegisterComponents('Misc2', [TWhatIsThis]);
end;

end.
